/*
 * Copyright (c) 2010,2012 Gražvydas Ignotas
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/XKBlib.h>
#include <X11/XF86keysym.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <signal.h>
#include <linux/kd.h>

#define PFX "op_runfbapp: "

static struct termios g_kbd_termios_saved;
static int g_kbdfd = -1;
static int g_have_x;
static pid_t g_child_pid;
static pthread_cond_t g_start_cond = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t g_start_mutex = PTHREAD_MUTEX_INITIALIZER;

static void signal_main_thread(void)
{
	pthread_mutex_lock(&g_start_mutex);
	pthread_cond_signal(&g_start_cond);
	pthread_mutex_unlock(&g_start_mutex);
}

static Cursor transparent_cursor(Display *display, Window win)
{
	Cursor cursor;
	Pixmap pix;
	XColor dummy;
	char d = 0;

	memset(&dummy, 0, sizeof(dummy));
	pix = XCreateBitmapFromData(display, win, &d, 1, 1);
	cursor = XCreatePixmapCursor(display, pix, pix,
			&dummy, &dummy, 0, 0);
	XFreePixmap(display, pix);
	return cursor;
}

static void x11h_wait_vmstate(Display *display)
{
	Atom wm_state = XInternAtom(display, "WM_STATE", False);
	XEvent evt;
	int i;

	for (i = 0; i < 20; i++) {
		while (XPending(display)) {
			XNextEvent(display, &evt);
			if (evt.type == PropertyNotify && evt.xproperty.atom == wm_state)
				return;
		}
		usleep(200000);
	}

	fprintf(stderr, "timeout waiting for wm_state change\n");
}

static int x11h_minimize(Display *display, Window window)
{
	XSetWindowAttributes attributes;
	int screen = DefaultScreen(display);
	int display_width, display_height;
	XWMHints wm_hints;
	XEvent evt;

	XWithdrawWindow(display, window, screen);

	attributes.override_redirect = False;
	XChangeWindowAttributes(display, window,
		CWOverrideRedirect, &attributes);

	wm_hints.flags = StateHint;
	wm_hints.initial_state = IconicState;
	XSetWMHints(display, window, &wm_hints);

	XMapWindow(display, window);

	while (XNextEvent(display, &evt) == 0)
	{
		// printf("m event %d\n", evt.type);
		switch (evt.type)
		{
			case FocusIn:
				goto out;
			default:
				break;
		}
	}

out:
	XWithdrawWindow(display, window, screen);

	// must wait for some magic vmstate property change before setting override_redirect
	x11h_wait_vmstate(display);

	attributes.override_redirect = True;
	XChangeWindowAttributes(display, window,
		CWOverrideRedirect, &attributes);

	// fixup window after resize on override_redirect loss
	display_width = DisplayWidth(display, screen);
	display_height = DisplayHeight(display, screen);
	XMoveResizeWindow(display, window, 0, 0, display_width, display_height);

	XMapWindow(display, window);
	XGrabKeyboard(display, window, False, GrabModeAsync, GrabModeAsync, CurrentTime);
	XkbSetDetectableAutoRepeat(display, 1, NULL);

	// we don't know when event dispatch will be called, so sync now
	XSync(display, False);

	return 0;
}

static void *x11_handler(void *arg)
{
	unsigned int display_width, display_height;
	XSetWindowAttributes attributes;
	const char *window_title = arg;
	Window win;
	XEvent evt;
	Display *display;
	int screen;

	display = XOpenDisplay(NULL);
	if (display == NULL)
	{
		fprintf(stderr, PFX "(not hiding X): Can't open display: %s\n",
			XDisplayName(NULL));
		signal_main_thread();
		return NULL;
	}
	g_have_x = 1;

	screen = DefaultScreen(display);

	display_width = DisplayWidth(display, screen);
	display_height = DisplayHeight(display, screen);

	win = XCreateSimpleWindow(display,
			RootWindow(display, screen),
			0, 0, display_width, display_height, 0,
			BlackPixel(display, screen),
			BlackPixel(display, screen));

	attributes.override_redirect = True;
	attributes.cursor = transparent_cursor(display, win);
	XChangeWindowAttributes(display, win, CWOverrideRedirect | CWCursor, &attributes);

	XStoreName(display, win, window_title);
	XSelectInput(display, win, ExposureMask | FocusChangeMask
		| KeyPressMask | KeyReleaseMask | PropertyChangeMask);
	XMapWindow(display, win);
	XGrabKeyboard(display, win, False, GrabModeAsync, GrabModeAsync, CurrentTime);
	XkbSetDetectableAutoRepeat(display, 1, NULL);
	XSync(display, False);

	while (1)
	{
		XNextEvent(display, &evt);

		switch (evt.type) {
		case Expose:
			signal_main_thread();
			while (XCheckTypedEvent(display, Expose, &evt))
				;
			break;

		case KeyPress:
			if (XLookupKeysym(&evt.xkey, 0) == XF86XK_MenuKB
			    && g_child_pid != 0) {
				kill(g_child_pid, SIGSTOP);
				x11h_minimize(display, win);
				kill(g_child_pid, SIGCONT);
			}
			break;
		}
	}

	return NULL;
}

static void hidecon_start(void)
{
	struct termios kbd_termios;
	int mode;

	g_kbdfd = open("/dev/tty", O_RDWR);
	if (g_kbdfd == -1) {
		perror(PFX "open /dev/tty");
		return;
	}

	if (ioctl(g_kbdfd, KDGETMODE, &mode) == -1) {
		perror(PFX "(not hiding FB): KDGETMODE");
		goto fail;
	}

	if (tcgetattr(g_kbdfd, &kbd_termios) == -1) {
		perror(PFX "tcgetattr");
		goto fail;
	}

	g_kbd_termios_saved = kbd_termios;
	kbd_termios.c_lflag &= ~(ICANON | ECHO); // | ISIG);
	kbd_termios.c_iflag &= ~(ISTRIP | IGNCR | ICRNL | INLCR | IXOFF | IXON);
	kbd_termios.c_cc[VMIN] = 0;
	kbd_termios.c_cc[VTIME] = 0;

	if (tcsetattr(g_kbdfd, TCSAFLUSH, &kbd_termios) == -1) {
		perror(PFX "tcsetattr");
		goto fail;
	}

	if (ioctl(g_kbdfd, KDSETMODE, KD_GRAPHICS) == -1) {
		perror(PFX "KDSETMODE KD_GRAPHICS");
		tcsetattr(g_kbdfd, TCSAFLUSH, &g_kbd_termios_saved);
		goto fail;
	}

	return;

fail:
	close(g_kbdfd);
	g_kbdfd = -1;
}

static void hidecon_end(void)
{
	if (g_kbdfd < 0)
		return;

	if (ioctl(g_kbdfd, KDSETMODE, KD_TEXT) == -1)
		perror(PFX "KDSETMODE KD_TEXT");

	if (tcsetattr(g_kbdfd, TCSAFLUSH, &g_kbd_termios_saved) == -1)
		perror(PFX "tcsetattr");

	close(g_kbdfd);
	g_kbdfd = -1;
}

static void do_exec(char * const argv[])
{
	int ret, status;
	pid_t pid;

	pid = fork();
	if (pid == -1) {
		perror(PFX "fork");
		return;
	}

	if (pid == 0) {
		/* child */
		execvp(argv[0], argv);
		perror(PFX "execvp");
		exit(1);
	}

	g_child_pid = pid;

	ret = waitpid(pid, &status, 0);
	if (ret < 0)
		perror(PFX "waitpid");
}

int main(int argc, char *argv[])
{
	pthread_t tid;
	char *name;
	int ret;

	if (argc < 2) {
		printf("usage:\n%s <app> [arg]..\n", argv[0]);
		return 1;
	}

	pthread_mutex_lock(&g_start_mutex);

	name = strrchr(argv[1], '/');
	if (name == NULL)
		name = argv[1];
	else
		name++;

	ret = pthread_create(&tid, NULL, x11_handler, name);
	if (ret != 0) {
		fprintf(stderr, PFX "pthread_create: %d\n", ret);
		return 1;
	}
	pthread_detach(tid);

	pthread_cond_wait(&g_start_cond, &g_start_mutex);
	pthread_mutex_unlock(&g_start_mutex);

	if (!g_have_x)
		hidecon_start();

	do_exec(argv + 1);

	if (!g_have_x)
		hidecon_end();

	/* XXX: maybe stop the X thread nicely? */

	return 0;
}

